import 'dart:convert';
import 'dart:math';
import 'package:easyflu/easyflu.dart';
import 'package:encrypt/encrypt.dart';
import 'package:flutter/foundation.dart' as platForm;
import 'package:easyflu/config/const.dart';
import 'package:easyflu/utils/object_util.dart';
import 'package:easyflu/utils/encrypt/rsa_util.dart';
import 'package:hive/hive.dart';
import 'package:pointycastle/export.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'encrypt/encrypt_util.dart';


enum StorageType{
  SP,
  HIVE
}


class StorageUtil {
  static StorageUtil _instance = StorageUtil._internal();
  late SharedPreferences _prefs;
  late Box _realEncryptBox;
  late StorageType _storageType;
  late Encrypter _realEncrypt;
  late String _hiveName;
  final _aesIv = IV.fromLength(16);
  late bool _encrypt;
  String _randomKey = EncryptUtil.encodeMd5("RANDOM_KEY");
  String _rsaKey = EncryptUtil.encodeMd5("RSA_KEY");
  String _defHiveName = EncryptUtil.encodeMd5("CacheStorage");
  factory StorageUtil() => _instance;


  // 私有构造函数
  StorageUtil._internal();


  Future init({
    bool encrypt = true,
    StorageType storageType = StorageType.HIVE,
    String? hiveSubDir,
    String hiveName = ""
  }) async{
    _storageType = storageType;
    _encrypt = encrypt;
    _hiveName = ObjectUtil.isEmpty(hiveName)? _defHiveName : hiveName;
    _prefs = await SharedPreferences.getInstance();


    ///先获取rsa密钥
    if(encrypt){

      /**
       * 基本加密逻辑
       *
       * 1.生成一个随机数，使用公钥秘钥进行加密,公钥秘钥使用salsa加密
       * 2.使用这个随机数来进行aes加解密
       * 3.key使用salsa进行加密混淆
       */

      ///第一层加密
      final key = Key.fromUtf8(defPassword);
      final iv = IV.fromLength(8);
      final salsaEncrypt = Encrypter(Salsa20(key));
      RSAPrivateKey rsaPrivateKey;
      RSAPublicKey rsaPublicKey;
      late String realRandomKey;
      String publicKey;
      String privateKey;
      String publicPrivateKey;
      var aesKey;


      if(_prefs.containsKey(_randomKey)){
        publicPrivateKey = _prefs.getString(_rsaKey) ?? "";
        publicPrivateKey = salsaEncrypt.decrypt64(publicPrivateKey, iv: iv);

        publicKey = publicPrivateKey.split(rsaSplit)[0];
        privateKey = publicPrivateKey.split(rsaSplit)[1];
        rsaPublicKey = RsaUtil.rsaPublicKeyFromPem(publicKey);
        rsaPrivateKey = RsaUtil.rsaPrivateKeyFromPem(privateKey);
        String encryptRandomKey = _prefs.getString(_randomKey) ?? "";

        ///解密randomKey
        realRandomKey = RsaUtil.rsaDecrypt(encryptRandomKey, rsaPrivateKey);

      }else{
        ///未进行过加密
        var pair = RsaUtil.generateRSAKeyPair();
        rsaPrivateKey = pair.privateKey as RSAPrivateKey;
        rsaPublicKey = pair.publicKey as RSAPublicKey;

        privateKey = RsaUtil.encodeRSAPrivateKeyToPem(rsaPrivateKey);
        publicKey = RsaUtil.encodeRSAPublicKeyToPem(rsaPublicKey);

        ///保存到本地,先进行salsa20加密
        publicPrivateKey = "$publicKey$rsaSplit$privateKey";
        publicPrivateKey = salsaEncrypt.encrypt(publicPrivateKey, iv: iv).base64;
        _prefs.setString(_rsaKey, publicPrivateKey);

        ///保存随机数
        realRandomKey = _generatePassword();
        String encryptRandomKey = RsaUtil.rsaEncrypt(realRandomKey, rsaPublicKey);
        _prefs.setString(_randomKey, encryptRandomKey);
      }

      aesKey = Key.fromUtf8(realRandomKey);
      _realEncrypt = Encrypter(AES(aesKey));

      if(_storageType == StorageType.HIVE){
        _realEncryptBox = await Hive.openBox(_hiveName, encryptionCipher: HiveAesCipher(base64Url.decode(realRandomKey)));
      }
    }else{
      _realEncryptBox = await Hive.openBox(_hiveName);
    }

  }



  ///强密码生成器
  String _generatePassword({
    bool letter = true,
    bool isNumber = true,
    bool isSpecial = true,
    int length = 32,
  }) {
    final letterLowerCase = "abcdefghijklmnopqrstuvwxyz";
    final letterUpperCase = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    final number = '0123456789';
    final special = '@#%^*>\$@?/[]=+';

    String chars = "";
    if (letter) chars += '$letterLowerCase$letterUpperCase';
    if (isNumber) chars += '$number';
    if (isSpecial) chars += '$special';

    return List.generate(length, (index) {
      final indexRandom = Random.secure().nextInt(chars.length);
      return chars [indexRandom];
    }).join('');
  }


  /// ********************************************统一处理*************************************************************///

  ///处理key，value
  String _handleKey(String key, {bool isEncrypt = false}){
    String resultKey = key;
    if(_encrypt){
      ///是否加密
      if(isEncrypt){
        resultKey = _realEncrypt.encrypt(key, iv: _aesIv).base64;
      }else{
        resultKey = _realEncrypt.decrypt64(key, iv: _aesIv);
      }
    }
    return resultKey;
  }


  String _handleValue(String value, {bool isEncrypt = false}){
    String resultValue = value;
    if(_encrypt){
      if(isEncrypt){
        resultValue = _realEncrypt.encrypt(value, iv: _aesIv).base64;
      }else{
        resultValue = _realEncrypt.decrypt64(value, iv: _aesIv);
      }
    }
    return resultValue;
  }


  Future _put(String key, String value) async{
    String resultKey = await _handleKey(key);
    String resultValue = await _handleValue(value);
    if(_storageType == StorageType.HIVE){
      return _realEncryptBox.put(resultKey, resultValue);
    }else if(_storageType == StorageType.SP){
      return _prefs.setString(resultKey, resultValue);
    }
  }


  String _get(String key, String defValue){
    String resultKey = _handleKey(key);
    String result = defValue;
    if(containsKey(resultKey)){
      if(_storageType == StorageType.HIVE){
        result = _handleValue(_realEncryptBox.get(resultKey));
      }else if(_storageType == StorageType.SP){
        result = _handleValue(_prefs.getString(resultKey) ?? defValue);
      }
    }
    return result;
  }




  /// ********************************************************put************************************************************************************///

  Future<void> putString(String key, String value) {
    return _put(key, value);
  }

  Future<void> putBool(String key, bool value) {
    return _put(key, value.toString());
  }


  Future<void> putInt(String key, int value) {
    return _put(key, value.toString());
  }


  Future<void> putDouble(String key, double value) {
    return _put(key, value.toString());
  }



  Future<void> putObject(String key, dynamic object){
    return _put(key, jsonEncode(object));
  }


  Future<void> putList(String key, List list) {
    list = list.map((e) => json.encode(e)).toList();
    return _put(key, json.encode(list));
  }



  /// *********************************************************get***********************************************************************************///

  String getString(String key, {String defValue: ''}){
    return _get(key, defValue);
  }


  bool getBool(String key, {bool defValue: false}) {
    return _get(key, defValue.toString()) as bool;
  }


  int getInt(String key, {int defValue: 0}) {
    return int.tryParse(_get(key, defValue.toString())) ?? defValue;
  }


  double getDouble(String key, {double defValue: 0.0}) {
    return double.tryParse(_get(key, defValue.toString())) ?? defValue;
  }



  T? getObject<T>(String key, T formMap(Map map)){
    String jsonResult = _get(key, "");
    if(ObjectUtil.isNotEmpty(jsonResult)){
      return formMap(json.decode(jsonResult));
    }
    return null;
  }


  List<T> getList<T>(String key, T formMap(Map v)){
    String jsonResult = _get(key, "");
    if(ObjectUtil.isNotEmpty(jsonResult)){
      List list = json.decode(jsonResult);
      return list.map((e){
        if(e is String){
          return formMap(json.decode(e));
        }
        return formMap(e);
      }).toList();
    }
    return [];
  }


  /// ***********************************************************key信息*********************************************************************************///

  bool haveKey(String key) {
    bool haveKey = false;
    if(_storageType == StorageType.HIVE){
      haveKey = _realEncryptBox.containsKey(key);
    }else if(_storageType == StorageType.SP){
      if(_encrypt){
        haveKey = _prefs.containsKey(_realEncrypt.encrypt(key).base64);
      }else{
        haveKey = _prefs.containsKey(key);
      }
    }
    return haveKey;
  }



  bool containsKey(String key){
    if(_storageType == StorageType.HIVE){
      return _realEncryptBox.containsKey(key);
    }else if(_storageType == StorageType.SP){
      return _prefs.containsKey(key);
    }
    return false;
  }



  /// ***********************************************************清除*********************************************************************************///
  Future remove(String key) {
    if(_storageType == StorageType.HIVE){
      return _realEncryptBox.delete(key);
    }else if(_storageType == StorageType.SP){
      if(_encrypt){
        return _prefs.remove(_realEncrypt.encrypt(key).base64);
      }else{
        return _prefs.clear();
      }
    }
    return Future.value(false);
  }


  /// clear.
  Future clear() {
    if(_storageType == StorageType.HIVE){
      return _realEncryptBox.clear();
    }else if(_storageType == StorageType.SP){
      return _prefs.clear();
    }
    return Future.value(false);
  }
}
